%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CROSS ENTROPY METHOD AND EVOLUTIONARY PARTICLE SWARM OPTIMIZATION (in MATLAB)
%
% Leonel Carvalho (leonel.m.carvalho@inesctec.pt)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gbest, gbestFitVector, gbestObjVector, otherParameters] = ...
    CEEPSO(iRun, ceepsoParameters,caseStudyData,otherParameters,lowerBounds,upperBounds,initialSolution)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% General Initializations
dim = ceepsoParameters.dim;
fnc = ceepsoParameters.fnc;
gbestFitVector = nan(1,ceepsoParameters.max_fit_eval / ceepsoParameters.pop_size);
gbestObjVector = nan(2,ceepsoParameters.max_fit_eval / ceepsoParameters.pop_size);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CE Initializations
n_samples = ceepsoParameters.pop_size;
ro = 0.1;
sigma = 0.85;
beta = 0.95;
q = 5;
indexRo = floor( ( 1 - ro ) * n_samples );
% indexRo = n_samples - 10;
H = zeros( 1, n_samples );
pos = zeros( n_samples, dim );
gammaIter = zeros( 1, 1e6 );
x_min = zeros( 1, dim );
x_max = zeros( 1, dim );
n_trials = ones( 1, dim );
meanPar = zeros( 1, dim );
stdPar = zeros( 1, dim );
for i = 1 : dim
    x_min( 1, i ) = lowerBounds( 1, i );
    x_max( 1, i ) = upperBounds( 1, i );
    meanPar( 1, i ) = x_min( 1, i ) + rand() * ( x_max( 1, i ) - x_min( 1, i ) );
    stdPar( 1, i ) = ( x_max( 1, i ) - x_min( 1, i ) ) * 5;
end
iter = 1;
memMeanPar( iter, : ) = meanPar;
memStdPar( iter, : ) = stdPar;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% EPSO Initializations
global epso_par;
epso_par.pop_size = ceepsoParameters.pop_size;
epso_par.mutation_rate = ceepsoParameters.mutation_rate;
epso_par.communication_probability = ceepsoParameters.communication_probability;
% Weights matrix
% 1 - inertia
% 2 - memory
% 3 - cooperation
% 4 - perturbation
weights = rand( epso_par.pop_size, 4 );
epso_par.init = 1;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fitness Function Initializations
global ff
ff.num_eval = 0;
ff.coeff = ones( 1, 4 );
ff.weights = zeros( n_samples, 4 );
ff.i_sample = 0;
gbestval = 1e30;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%
while ff.num_eval < ceepsoParameters.max_fit_eval
    %%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Run CE
    if ff.num_eval <= ceepsoParameters.max_fit_eval_CE
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Generate population
        for i = 1 : dim
            tmp_stat = normrnd( meanPar( i ), stdPar( i ), 1, n_samples );
            tmp_stat( tmp_stat < x_min( 1, i ) ) = x_min( 1, i );
            tmp_stat( tmp_stat > x_max( 1, i ) ) = x_max( 1, i );
            tmp_eval = tmp_stat;
            pos( :, i ) = tmp_eval;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate population
        [fit, obj, pos, otherParameters]=feval(fnc,pos,caseStudyData,otherParameters);
        ff.num_eval = ff.num_eval + ceepsoParameters.pop_size;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Performs CE correction
       
        % Sorts states per evaluation
        [ Saux, Iaux ] = sort( fit, 1, 'descend' );
        H = zeros( 1, n_samples );
        for i = 1 : n_samples
            if i <= indexRo
                H( Iaux( i ) ) = 0;
            else
                H( Iaux( i ) ) = 1;
            end
        end
        
        % Computes gammaIter
        gamma = Saux( indexRo );
        gammaIter( 1, iter ) = gamma;
        
        % Applies CE correction
        den = sum( H );
        for i = 1 : dim
            num = H * pos( :, i );
            tmpMean = 0;
            if den > 0
                tmpMean = ( num / den ) * ( 1 / n_trials( 1, i ) );
            end
            num = H * ( ( pos( :, i ) - tmpMean ) .* ( pos( :, i ) - tmpMean ) );
            tmpStd = 0;
            if den > 0
                tmpStd = sqrt( num / den ) * ( 1 / n_trials( 1, i ) );
            end
            
            meanPar( i ) = tmpMean * sigma + meanPar( i ) * ( 1 - sigma );
            
            % Update smoothing parameters
            sigmaStd = beta - beta * ( 1 - 1 / iter )^q;
            stdPar( i ) = tmpStd * sigmaStd + stdPar( i ) * ( 1 - sigmaStd );
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Update weights fitness function
        tmpMat = ff.weights( 1:ff.i_sample, : );
        tmp = max( tmpMat );
        ff.coeff = 10 .^ round( log10( max( tmp )./ tmp )  );
        ff.coeff( isinf( ff.coeff ) ) = 1;
        ff.i_sample = 0;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        memMeanPar( iter, : ) = meanPar;
        memStdPar( iter, : ) = stdPar;
    
    else
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Run EPSO
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %  Initialize EPSO population
        if epso_par.init == 1
            Xmin = x_min;
            Xmax = x_max;
            Vmin = -Xmax + Xmin;
            Vmax = -Vmin;
            vel = zeros( epso_par.pop_size, dim );
            for i = 1 : epso_par.pop_size
                vel( i, : ) = Vmin + ( Vmax - Vmin ) .* rand( 1, dim );
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Evaluate population
            [fit, obj, pos, otherParameters]=feval(fnc,pos,caseStudyData,otherParameters);
            ff.num_eval = ff.num_eval + ceepsoParameters.pop_size;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % UPDATE individual best
            % Individual best position ever of the particles of current population
            myBestPos = pos;
            % Fitness of the individual best position ever of the particles of current population
            myBestPosFit = fit;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            epso_par.init = 0;
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Update global best
            [ tmpgbestval, indexbest ] = min( fit );
            if( tmpgbestval < gbestval )
                gbestval = tmpgbestval;
                objbestval = obj( indexbest, : );
                gbest = pos( indexbest, : );
                otherParameters.idBestParticle = indexbest;
                otherParameters.pfFinal = otherParameters.pfDB(:,indexbest);
                otherParameters.genCostsFinal = otherParameters.genCosts(indexbest,:);
                otherParameters.loadDRcostsFinal = otherParameters.loadDRcosts(indexbest,:);
                otherParameters.v2gChargeCostsFinal = otherParameters.v2gChargeCosts(indexbest,:);
                otherParameters.v2gDischargeCostsFinal = otherParameters.v2gDischargeCosts(indexbest,:);
                otherParameters.storageChargeCostsFinal = otherParameters.storageChargeCosts(indexbest,:);
                otherParameters.storageDischargeCostsFinal = otherParameters.storageDischargeCosts(indexbest,:);
                otherParameters.stBalanceFinal = otherParameters.stBalance(indexbest,:,:);
                otherParameters.v2gBalanceFinal = otherParameters.v2gBalance(indexbest,:,:);
                otherParameters.pensVoltageUFinal = otherParameters.pensVoltageU(indexbest,:);
                otherParameters.pensVoltageLFinal = otherParameters.pensVoltageL(indexbest,:);
                otherParameters.pensMaxSLinesFinal = otherParameters.pensMaxSLines(indexbest,:);
                otherParameters.penSlackBusFinal = otherParameters.penSlackBus(indexbest,:);
            end
            gbestFitVector( 1, iter ) = gbestval;
            gbestObjVector( :, iter ) = objbestval;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % COPY CURRENT population
        copy_pos = pos;
        copy_vel = vel;
        copy_weights = weights;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Apply EPSO movement rule
        for i = 1 : epso_par.pop_size
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Mutate weights of the particles of the COPIED population
            copy_weights( i, : ) = MUTATE_WEIGHTS( weights( i, : ), epso_par.mutation_rate );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new velocity for the particles of the COPIED population
            copy_vel( i, : ) = EPSO_COMPUTE_NEW_VEL( dim, copy_pos( i, : ), myBestPos( i, : ), gbest, ...
                copy_vel( i, : ), Vmin, Vmax, copy_weights( i, : ), epso_par.communication_probability );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new position for the particles of the COPIED population
            [ copy_pos( i, : ), copy_vel( i, : ) ] = COMPUTE_NEW_POS( copy_pos( i, : ), copy_vel( i, : ) );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new velocity for the particles of the CURRENT population
            vel( i, : ) = EPSO_COMPUTE_NEW_VEL( dim, pos( i, : ), myBestPos( i, : ), gbest, ...
                vel( i, : ), Vmin, Vmax, weights( i, : ), epso_par.communication_probability );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new position for the particles of the CURRENT population
            [ pos( i, : ), vel( i, : ) ] = COMPUTE_NEW_POS( pos( i, : ), vel( i, : ) );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Enforce search space limits of the COPIED population
        [ copy_pos, copy_vel ] = ENFORCE_POS_LIMITS( dim, epso_par.pop_size, copy_pos, Xmin, Xmax, copy_vel, Vmin, Vmax );
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Enforce search space limits of the CURRENT population
        [ pos, vel ] = ENFORCE_POS_LIMITS( dim, epso_par.pop_size, pos, Xmin, Xmax, vel, Vmin, Vmax );
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate the COPIED population
        [ copy_fit, copy_obj, copy_pos, copy_otherParameters ] = feval(fnc,copy_pos,caseStudyData,otherParameters);
        ff.num_eval = ff.num_eval + ceepsoParameters.pop_size;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate the CURRENT population
        [ fit, obj, pos, otherParameters ] = feval(fnc,pos,caseStudyData,otherParameters);
        ff.num_eval = ff.num_eval + ceepsoParameters.pop_size;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % CREATE NEW population to replace CURRENT population
        selParNewSwarm = ( copy_fit < fit );
        for i = 1 : epso_par.pop_size
            if selParNewSwarm( i )
                fit( i ) = copy_fit( i );
                obj( i , : ) = copy_obj( i, : );
                pos( i, : ) = copy_pos( i, : );
                vel( i, : ) = copy_vel( i, : );
                weights( i, : ) = copy_weights( i, : );
                otherParameters.pfDB(:,i) = copy_otherParameters.pfDB(:,i);
                otherParameters.genCosts(i,:) = copy_otherParameters.genCosts(i,:);
                otherParameters.loadDRcosts(i,:) = copy_otherParameters.loadDRcosts(i,:);
                otherParameters.v2gChargeCosts(i,:) = copy_otherParameters.v2gChargeCosts(i,:);
                otherParameters.v2gDischargeCosts(i,:) = copy_otherParameters.v2gDischargeCosts(i,:);
                otherParameters.storageChargeCosts(i,:) = copy_otherParameters.storageChargeCosts(i,:);
                otherParameters.storageDischargeCosts(i,:) = copy_otherParameters.storageDischargeCosts(i,:);
                otherParameters.stBalance(i,:,: )= copy_otherParameters.stBalance(i,:,:);
                otherParameters.v2gBalance(i,:,:) = copy_otherParameters.v2gBalance(i,:,:);
                otherParameters.pensVoltageU(i,:) = copy_otherParameters.pensVoltageU(i,:);
                otherParameters.pensVoltageL(i,:) = copy_otherParameters.pensVoltageL(i,:);
                otherParameters.pensMaxSLines(i,:) = copy_otherParameters.pensMaxSLines(i,:);
                otherParameters.penSlackBus(i,:) = copy_otherParameters.penSlackBus(i,:);
            end
            if fit( i ) < myBestPosFit( i )
                myBestPos( i, : ) = pos( i, : );
                myBestPosFit( i ) = fit( i );
            end
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		
		%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		% Add random velocity 
		if rem( ff.num_eval, 10000 ) == 0 && ff.num_eval > ceepsoParameters.max_fit_eval_CE
			for i = 1 : epso_par.pop_size          
            vel( i, : ) = ( Vmin + ( Vmax - Vmin ) .* rand( 1, dim ) ) * 0.25;
			end
		end
		%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Update global best
    [ tmpgbestval, indexbest ] = min( fit );
    if( tmpgbestval < gbestval )
        gbestval = tmpgbestval;
        objbestval = obj( indexbest, : );
        gbest = pos( indexbest, : );
        otherParameters.idBestParticle = indexbest;
        otherParameters.pfFinal = otherParameters.pfDB(:,indexbest);
        otherParameters.genCostsFinal = otherParameters.genCosts(indexbest,:);
        otherParameters.loadDRcostsFinal = otherParameters.loadDRcosts(indexbest,:);
        otherParameters.v2gChargeCostsFinal = otherParameters.v2gChargeCosts(indexbest,:);
        otherParameters.v2gDischargeCostsFinal = otherParameters.v2gDischargeCosts(indexbest,:);
        otherParameters.storageChargeCostsFinal = otherParameters.storageChargeCosts(indexbest,:);
        otherParameters.storageDischargeCostsFinal = otherParameters.storageDischargeCosts(indexbest,:);
        otherParameters.stBalanceFinal = otherParameters.stBalance(indexbest,:,:);
        otherParameters.v2gBalanceFinal = otherParameters.v2gBalance(indexbest,:,:);
        otherParameters.pensVoltageUFinal =  otherParameters.pensVoltageU(indexbest,:);
        otherParameters.pensVoltageLFinal = otherParameters.pensVoltageL(indexbest,:);
        otherParameters.pensMaxSLinesFinal = otherParameters.pensMaxSLines(indexbest,:);
        otherParameters.penSlackBusFinal = otherParameters.penSlackBus(indexbest,:);
    end    
    gbestFitVector( 1, iter ) = gbestval;
    gbestObjVector( :, iter ) = objbestval;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Update iteration
    iter = iter + 1;
    fprintf('num fit eval = %d ; best fit = %.2f \n',  ff.num_eval, gbestval );
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

end

end