% DIFFERENTIAL SEARCH ALGORITHM (DSA) (in MATLAB)
% STANDARD VERSION of DSA (16.July.2013)
%
% modified by GECAD
%
% usage : > ds(method,fnc,mydata,popsize,dim,low,up,maxcycle)
%
% method
%--------------------------------------------------------------
% 1: Bijective DSA (B-DSA)
% 2: Surjective DSA (S-DSA)
% 3: Elitist DSA (strategy 1) (E1-DSA)
% 4: Elitist DSA (strategy 2) (E2-DSA)
% if method=[. . . ...],   Hybrid-DSA (H-DSA)
%--------------------------------------------------------------
% example :
% ds(1,'circlefit',mydata,10,3,-10,10,2000) ; % B-DSA
% ds(2,'circlefit',mydata,10,3,-10,10,2000) ; % S-DSA
% ds(3,'circlefit',mydata,10,3,-10,10,2000) ; % E1-DSA
% ds(4,'circlefit',mydata,10,3,-10,10,2000) ; % E2-DSA
% ds([1 2],'circlefit',mydata,10,3,-10,10,2000) ; % Hybrid-DSA, in this case B-DSA and S-DSA are hybridized.
%--------------------------------------------------------------
% Please cite this article as;
% P.Civicioglu, "Transforming geocentric cartesian coordinates to geodetic coordinates by using differential search algorithm",  Computers & Geosciences, 46 (2012), 229-247.
% P.Civicioglu, "Understanding the nature of evolutionary search algorithms", Additional technical report for the project of 110Y309-Tubitak,2013.
%
% (Basic/simple version can be found in matlab file exchange)
%
%
%--------------------------------------------------------------
%{
19.March.2013
Copyright Notice
Copyright (c) 2012, Pinar Civicioglu
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution
      
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%}
% has been modified by GECAD for the WG-competition 2017

function [globalminimizer, fitMaxVector, objMaxVector, otherParameters] = ...
    DSA(dsaParameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit,initialSolution)
% size_of_superorganism ; size of population.
% size_of_one_clan ; size of problem dimension (1,2,3,...,d), where each clan (i.e. sub-superorganism) includes d-individuals.
% mydata ; additional parameters for objective function, use mydata=[], if it is not needed. See Best-Fit Circle example (circlefit.m) for usage of mydata.

%Initialization

fnc=dsaParameters.fnc;
method=dsaParameters.method;
%R=dsaParameters.scale_factor;
p1_multiplier=dsaParameters.p1_multiplier;
p2_multiplier=dsaParameters.p2_multiplier;
size_of_superorganism=dsaParameters.nParticles;
epoch=dsaParameters.maxIterations;
minEpochs=dsaParameters.minIterations;
threshold=dsaParameters.threshold;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% size_of_one_clan -> PSO -> nVariables / size_of_superorganism -> PSO -> nParticles
size_of_one_clan = numel(up_habitat_limit);
dsaParameters.nVariables=size_of_one_clan;

% pre-allocation of loop variables
fitMaxVector = nan(1,epoch);
objMaxVector = nan(2,epoch);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% control of habitat limits
if numel(low_habitat_limit)==1,
    low_habitat_limit=low_habitat_limit*ones(1,size_of_one_clan);
    up_habitat_limit=up_habitat_limit*ones(1,size_of_one_clan);
end

% limit iterations by threshold
fitIterationGap = inf;
noIterationsToGap = dsaParameters.noIterationsToGap;
epk = 1;

% JSOARES modification - vectorization
minPositionsMatrix=repmat(low_habitat_limit,size_of_superorganism,1);
maxPositionsMatrix=repmat(up_habitat_limit,size_of_superorganism,1);
dsaParameters.minPositionsMatrix=minPositionsMatrix;
dsaParameters.maxPositionsMatrix=maxPositionsMatrix;

% generate initial individuals, clans and superorganism.
superorganism=genpop(size_of_superorganism,size_of_one_clan,minPositionsMatrix,maxPositionsMatrix);
if nargin>5
    noInitialSolutions = size(initialSolution,1);
    superorganism(1:noInitialSolutions,:)=initialSolution;
end

superorganism=update(superorganism,minPositionsMatrix,maxPositionsMatrix);
% success of clans/superorganism
%fit_superorganism=feval(fnc,superorganism,mydata);
[fit_superorganism, obj_superorganism, superorganism, otherParameters]=feval(fnc,superorganism,caseStudyData, otherParameters);

[globalminimum,indexbest] = min(fit_superorganism);

% store other information
otherParameters.idBestParticle = indexbest;
otherParameters.pfFinal = otherParameters.pfDB(:,indexbest);
otherParameters.genCostsFinal = otherParameters.genCosts(indexbest,:);
otherParameters.loadDRcostsFinal = otherParameters.loadDRcosts(indexbest,:);
otherParameters.v2gChargeCostsFinal = otherParameters.v2gChargeCosts(indexbest,:);
otherParameters.v2gDischargeCostsFinal =otherParameters.v2gDischargeCosts(indexbest,:);
otherParameters.storageChargeCostsFinal = otherParameters.storageChargeCosts(indexbest,:);
otherParameters.storageDischargeCostsFinal = otherParameters.storageDischargeCosts(indexbest,:);
otherParameters.stBalanceFinal = otherParameters.stBalance(indexbest,:,:);
otherParameters.v2gBalanceFinal = otherParameters.v2gBalance(indexbest,:,:);
otherParameters.pensVoltageUFinal =  otherParameters.pensVoltageU(indexbest,:);
otherParameters.pensVoltageLFinal = otherParameters.pensVoltageL(indexbest,:);
otherParameters.pensMaxSLinesFinal = otherParameters.pensMaxSLines(indexbest,:);
otherParameters.penSlackBusFinal = otherParameters.penSlackBus(indexbest,:);
objMaxVector(:,1)=obj_superorganism(indexbest,:);

while epk<=epoch &&  fitIterationGap >= threshold
    
    
    % SETTING OF ALGORITHMIC CONTROL PARAMETERS
    % Trial-pattern generation strategy for morphogenesis; 'one-by-one morphogenesis'.
    % p1=0.0*rand;  % i.e.,  0.0 <= p1 <= 0.0
    % p2=0.0*rand;  % i.e.,  0.0 <= p2 <= 0.0
    
    % Trial-pattern generation strategy for morphogenesis; 'one-or-more morphogenesis'. (DEFAULT)
    p1=p1_multiplier*rand;  % i.e.,  0.0 <= p1 <= 0.3
    p2=p2_multiplier*rand;  % i.e.,  0.0 <= p2 <= 0.3
    
    %-------------------------------------------------------------------
    
    [direction,msg]=generate_direction(method(randi(numel(method))),superorganism,size_of_superorganism,fit_superorganism);
    
    map=generate_map_of_active_individuals(size_of_superorganism,size_of_one_clan,p1,p2);
    
    %-------------------------------------------------------------------
    % Recommended Methods for generation of Scale-Factor; R
    % R=4*randn;  % brownian walk
    % R=4*randg;  % brownian walk
    % R=lognrnd(rand,5*rand);  % brownian walk
    R=1./gamrnd(1,0.5);   % pseudo-stable walk
    % R=1/normrnd(0,5);    % pseudo-stable walk
    
    %-------------------------------------------------------------------
    
    % bio-interaction (morphogenesis)
    stopover=superorganism+(R.*map).*(direction-superorganism);
    
    % check signaling % JSOARES modification
    %[stopover,otherParameters] = checkSignaling(stopover,otherParameters,dsaParameters);
    
    % Boundary Control
    stopover=update(stopover,minPositionsMatrix,maxPositionsMatrix);
    
    % Selection-II
    
    [fit_stopover, obj_stopover, stopover, otherParameters]=feval(fnc,stopover,caseStudyData,otherParameters);
    ind=fit_stopover<fit_superorganism;
    fit_superorganism(ind)=fit_stopover(ind);
    superorganism(ind,:)=stopover(ind,:);
    
    % update results
    [globalminimum,indexbest]=min(fit_superorganism);
    globalminimizer=superorganism(indexbest,:);
    
    % store fitness evolution and obj fun evolution as well
    fitMaxVector(epk) = globalminimum;
    
    
    if ind(indexbest)
        % store other info
        otherParameters.idBestParticle = indexbest;
        otherParameters.pfFinal = otherParameters.pfDB(:,indexbest);
        otherParameters.genCostsFinal = otherParameters.genCosts(indexbest,:);
        otherParameters.loadDRcostsFinal = otherParameters.loadDRcosts(indexbest,:);
        otherParameters.v2gChargeCostsFinal = otherParameters.v2gChargeCosts(indexbest,:);
        otherParameters.v2gDischargeCostsFinal =otherParameters.v2gDischargeCosts(indexbest,:);
        otherParameters.storageChargeCostsFinal = otherParameters.storageChargeCosts(indexbest,:);
        otherParameters.storageDischargeCostsFinal = otherParameters.storageDischargeCosts(indexbest,:);
        otherParameters.stBalanceFinal = otherParameters.stBalance(indexbest,:,:);
        otherParameters.v2gBalanceFinal = otherParameters.v2gBalance(indexbest,:,:);
        otherParameters.pensVoltageUFinal =  otherParameters.pensVoltageU(indexbest,:);
        otherParameters.pensVoltageLFinal = otherParameters.pensVoltageL(indexbest,:);
        otherParameters.pensMaxSLinesFinal = otherParameters.pensMaxSLines(indexbest,:);
        otherParameters.penSlackBusFinal = otherParameters.penSlackBus(indexbest,:);
        objMaxVector(:,epk)= obj_stopover(indexbest,:);
    elseif epk>1
        objMaxVector(:,epk)=objMaxVector(:,epk-1);
    end
    
    
    globalminimum
    
    epk
  
    
    %%    objMaxVector (:,iIteration) = PSOparams.objMax;
    
    
    % export results - JSOARES DEACTIVATED 11-06-2014
    %     assignin('base','globalminimum',globalminimum);
    %     assignin('base','globalminimizer',globalminimizer);
    
    % fprintf('%s  | %5.0f   --->   %10.16f\n',msg,epk,globalminimum)
    
    if epk >= minEpochs
        fitIterationGap = abs(fitMaxVector(epk)-mean(fitMaxVector(epk-1:-1:epk-1-noIterationsToGap)));
    end
    
    epk=epk+1;
end


% JSOARES VERSION - I VECTORIZED THE CODE INSTEAD OF USING FOR
function pop=genpop(a,b,lowMatrix,upMatrix)
%pop=ones(a,b);
% for i=1:a
%     for j=1:b
%         pop(i,j)=rand*(up(j)-low(j))+low(j);
%     end
% end
pop=unifrnd(lowMatrix,upMatrix,a,b);

% JSOARES VERSION - I VECTORIZED THE CODE INSTEAD OF USING FOR
function p=update(p,lowMatrix,upMatrix)
%[popsize,dim]=size(p);
[idx] = find(p<lowMatrix);
p(idx)=lowMatrix(idx);
[idx] = find(p>upMatrix);
p(idx)=upMatrix(idx);

% for i=1:popsize
%     for j=1:dim
%         % first (standard)-method
%         if p(i,j)<low(j), if rand<rand, p(i,j)=rand*(up(j)-low(j))+low(j); else p(i,j)=low(j); end, end
%         if p(i,j)>up(j),  if rand<rand, p(i,j)=rand*(up(j)-low(j))+low(j); else p(i,j)=up(j); end, end
%
%         %{
%        %  second-method
%         if rand<rand,
%             if p(i,j)<low(j) || p(i,j)>up(j), p(i,j)=rand*(up(j)-low(j))+low(j); end
%         else
%             if p(i,j)<low(j), p(i,j)=low(j); end
%             if p(i,j)>up(j),  p(i,j)=up(j); end
%         end
%         %}
%     end
% end


function [direction,msg]=generate_direction(method,superorganism,size_of_superorganism,fit_superorganism);
switch method
    case 1,
        % BIJECTIVE DSA  (B-DSA) (i.e., go-to-rnd DSA);
        % philosophy: evolve the superorganism (i.e.,population) towards to "permuted-superorganism (i.e., random directions)"
        direction=superorganism(randperm(size_of_superorganism),:); msg=' B-DSA';
    case 2,
        % SURJECTIVE DSA (S-DSA) (i.e., go-to-good DSA)
        % philosophy: evolve the superorganism (i.e.,population) towards to "some of the random top-best" solutions
        ind=ones(size_of_superorganism,1);
        [null_,B]=sort(fit_superorganism);
        % JSOARES VERSION - TRY TO VECTORIZE
        ind=B(ceil(unifrnd(0,ceil(rand(size_of_superorganism,1)*size_of_superorganism))));
        % OLD
        %for i=1:size_of_superorganism, ind(i)=B(randi(ceil(rand*size_of_superorganism),1)); end;
        direction=superorganism(ind,:);  msg=' S-DSA';
    case 3,
        % ELITIST DSA #1 (E1-DSA) (i.e., go-to-best DSA)
        % philosophy: evolve the superorganism (i.e.,population) towards to "one of the random top-best" solution
        [null,jind]=sort(fit_superorganism); ibest=jind(ceil(rand*size_of_superorganism)); msg='E1-DSA';
        direction=repmat(superorganism(ibest,:),[size_of_superorganism 1]);
    case 4,
        % ELITIST DSA #2 (E2-DSA) (i.e., go-to-best DSA)
        % philosophy: evolve the superorganism (i.e.,population) towards to "the best" solution
        [null_,ibest]=min(fit_superorganism); msg='E2-DSA';
        direction=repmat(superorganism(ibest,:),[size_of_superorganism 1]);
end
return


function map=generate_map_of_active_individuals(size_of_superorganism,size_of_one_clan,p1,p2);
% strategy-selection of active/passive individuals
map=zeros(size_of_superorganism,size_of_one_clan);
if rand<rand,
    if rand<p1,
        % Random-mutation #1 strategy
        %JSOARES VERSION - VEC VERSION
        map = rand(size_of_superorganism,size_of_one_clan) < repmat(rand(size_of_superorganism,1),1,size_of_one_clan);
        % OLD
        %                 for i=1:size_of_superorganism
        %                     map(i,:)=rand(1,size_of_one_clan) < rand;
        %                 end
    else
        % Differential-mutation strategy
        %JSOARES VERSION - VECTORIZED
        map(:,randi(size_of_one_clan,size_of_superorganism,1))=1;
        % OLD
        %                 for i=1:size_of_superorganism
        %                     map(i,randi(size_of_one_clan))=1;
        %                 end
    end
else
    % Random-mutation #2 strategy
    %             temp=randi(size_of_one_clan,size_of_superorganism,ceil(p2*size_of_one_clan));
    %             temp=sub2ind(size(map),repmat([1:size_of_superorganism]',1,ceil(p2*size_of_one_clan)),temp);
    %             map(temp);
    % not working above
    for i=1:size_of_superorganism
        map(i,randi(size_of_one_clan,1,ceil(p2*size_of_one_clan)))=1;
    end
end
return